"""
Utility functions to support entity data testing
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import copy

from django.conf                    import settings
from django.http                    import QueryDict
from django.utils.http              import urlquote, urlunquote
from django.utils.html              import escape
from django.core.urlresolvers       import resolve, reverse
from django.template                import Context
from django.contrib.auth.models     import User

from utils.py3porting               import is_string, to_unicode

import annalist
from annalist.util                  import valid_id, extract_entity_id
from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist                       import layout

from annalist.models.annalistuser   import AnnalistUser

from annalist.views.uri_builder     import (
    uri_quote_param,
    uri_params, uri_with_params
    )
from annalist.views.fields.field_description    import FieldDescription
from annalist.views.fields.bound_field          import bound_field, get_entity_values
from annalist.views.fields.render_placement     import get_placement_classes
from annalist.views.form_utils.fieldchoice      import FieldChoice, update_choice_labels

from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )

#   -----------------------------------------------------------------------------
#
#   Message formatting functions
#
#   -----------------------------------------------------------------------------

def make_message(msg_format, coll_id="testcoll", type_id="testtype", **kwargs):
        """
        Combine supplied message format string with keyword parameters to build
        a message text.
        """
        msg_vals = dict(kwargs, coll_id=coll_id, type_id=type_id)
        msg_text = msg_format%msg_vals
        return msg_text

def make_quoted_message(msg_format, coll_id="testcoll", type_id="testtype", **kwargs):
        """
        Combine supplied message format string with keyword parameters to build
        a message text, then apply URL quoting to return the message as it will 
        appear as a URL parameter.
        """
        msg_text = make_message(msg_format, coll_id=coll_id, type_id=type_id, **kwargs)
        return uri_quote_param(msg_text)

#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def site_dir():
    return TestBaseDir + "/"

def collection_dir(coll_id="testcoll"):
    return (
        os.path.normpath(site_dir() + layout.SITE_COLL_PATH%{'id': coll_id})
        + "/"
        )

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

#   These all use the Django `reverse` function so they correspond to
#   the declared URI patterns.

def entitydata_list_url_query(viewname, kwargs, query_params, more_params):
    """
    Helper function for generatinglist URLs
    """
    list_url=reverse(viewname, kwargs=kwargs)
    return uri_with_params(list_url, query_params, more_params)

def site_view_url():
    return reverse("AnnalistSiteView")

def collection_view_url(coll_id="testcoll"):
    return reverse("AnnalistCollectionView", kwargs={'coll_id': coll_id})

def collection_edit_url(coll_id="testcoll"):
    return reverse("AnnalistCollectionEditView", kwargs={'coll_id': coll_id})

def collection_resource_url(coll_id="testcoll", resource_ref=layout.COLL_META_FILE):
    return reverse(
        "AnnalistCollectionResourceAccess", 
        kwargs={'coll_id': coll_id, 'resource_ref': resource_ref }
        )

def collection_entity_view_url(coll_id="testcoll", type_id=None, entity_id=None):
    """
    Return URL for entity belonging to some collection.

    This matches the 'annal:url' value that is stored in saved entity data.
    """
    return reverse(
        "AnnalistEntityAccessView", 
        kwargs={'coll_id': coll_id, 'type_id': type_id, 'entity_id': entity_id}
        )

def collection_entity_edit_url(
        coll_id="testcoll", type_id=None, entity_id=None, 
        view_id=None, action=None
        ):
    """
    Return URL for edit view of entity
    """
    viewname = ( 
        'AnnalistEntityNewView'             if action == "new" else
        'AnnalistEntityEditView'
        )
    args = {'action': action, 'coll_id': coll_id, 'view_id': view_id}
    if type_id:
        args.update({'type_id': type_id})
    if entity_id:
        args.update({'entity_id': entity_id})
    return reverse(viewname, kwargs=args)

def collection_entity_list_url(coll_id="testcoll", list_id=None, type_id=None, scope=None):
    """
    Return URL for entity list in collection some collection.

    url(r'^c/(?P<coll_id>\w{0,32})/l/$',
                            EntityDefaultListView.as_view(),
                            name='AnnalistEntityDefaultList'),
    url(r'^c/(?P<coll_id>\w{0,32})/l/(?P<list_id>\w{0,32})/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityGenericList'),
    url(r'^c/(?P<coll_id>\w{0,32})/l/(?P<list_id>\w{0,32})/(?P<type_id>\w{0,32})/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityGenericList'),

    """
    view = "AnnalistEntityDefaultList"
    args = {'coll_id': coll_id}
    if list_id is not None:
        view = "AnnalistEntityGenericList"
        args['list_id'] = list_id
    if type_id is not None:
        args['type_id'] = type_id
    return entitydata_list_url_query(view, args, None,
        { 'scope': scope
        })

def continuation_url_param(uri, prev_cont=None):
    if prev_cont:
        uri += "?" + prev_cont
    return "continuation_url=" + urlquote(uri, safe="/=!")

def confirm_delete_params(
        button_id="entity_delete", coll_id="testcoll", entity_id="entity1", search_for="", list_id=None, type_id="testtype"):
    vals = (
        { 'button_id':  button_id
        , 'entity_id':  entity_id
        , 'coll_id':    coll_id
        , 'list_id_':   "/l/"+list_id+"/" if list_id else "/d/"
        , 'type_id_':   type_id + '/' if type_id else ""
        , 'search_for': search_for
        })
    params = (
        """{"%(button_id)s": ["Delete"],"""+
        """ "entity_id": ["%(entity_id)s"],"""+
        """ "completion_url": ["/testsite/c/%(coll_id)s%(list_id_)s%(type_id_)s"],"""+
        """ "search_for": ["%(search_for)s"]}"""
        )%vals
    return params

#   -----------------------------------------------------------------------------
#
#   ----- Site data
#
#   -----------------------------------------------------------------------------

def site_title(template="%s"):
    return template%("Annalist data notebook test site")

#   -----------------------------------------------------------------------------
#
#   ----- Collection data
#
#   -----------------------------------------------------------------------------

def collection_value_keys():
    """
    Keys in collection data
    """
    return (
        [ '@type'
        , 'annal:id' 
        # , 'annal:type_id'
        , 'annal:url'
        , 'annal:uri'
        , 'rdfs:label'
        , 'rdfs:comment'
        ])

def collection_create_values(coll_id="testcoll", update="Collection"):
    """
    Entity values used when creating a collection entity
    """
    return (
        { 'rdfs:label':             "%s %s"%(update, coll_id)
        , 'rdfs:comment':           "Description of %s %s"%(update, coll_id)
        , 'annal:software_version': annalist.__version_data__
        })

def collection_values(coll_id, update="Collection", hosturi=TestHostUri):
    d = collection_create_values(coll_id, update=update).copy()
    d.update(
        { '@id':            "../"
        , '@type':          ["annal:Collection"]
        , 'annal:id':       coll_id
        , 'annal:type':     "annal:Collection"
        , 'annal:url':      collection_view_url(coll_id=coll_id)
        })
    return d

def collection_new_form_data(coll_id, update="Collection"):
    return (
        { "new":        "New collection"
        , "new_id":     coll_id
        , "new_label":  "%s %s"%(update, coll_id)
        })

def collection_remove_form_data(coll_id_list):
    return (
        { "remove":     "Remove selected"
        , "new_id":     ""
        , "new_label":  ""
        , "select":     coll_id_list
        })

#   -----------------------------------------------------------------------------
#
#   ----- Rendering support
#
#   -----------------------------------------------------------------------------

def render_select_options(name, label, opts, sel, placeholder=None):
    """
    Cf. `templates.field.annalist_edit_select.html`

    >>> print render_select_options("foo", "foo_label", ["aa", "bb", "cc"], "bb")
    <div class="row">
      <div class="small-10 columns view-value less-new-button">
        <select name="foo">
          <option value="aa">aa</option>
          <option value="bb" selected="selected">bb</option>
          <option value="cc">cc</option>
        </select>
      </div>
      <div class="small-2 columns view-value new-button left small-text-right">
        <button type="submit" 
                name="foo__new_edit" 
                value="New"
                title="Define new or edit foo_label">
          <span class="select-edit-button-text">+&#x270D;</span>
        </button>
      </div>
    </div>
    <BLANKLINE>
    >>> print render_select_options("foo", "foo_label", ["", "aa", "bb", "cc"], "", placeholder="(select)")
    <div class="row">
      <div class="small-10 columns view-value less-new-button">
        <select name="foo">
          <option value="" selected="selected">(select)</option>
          <option value="aa">aa</option>
          <option value="bb">bb</option>
          <option value="cc">cc</option>
        </select>
      </div>
      <div class="small-2 columns view-value new-button left small-text-right">
        <button type="submit" 
                name="foo__new_edit" 
                value="New"
                title="Define new or edit foo_label">
          <span class="select-edit-button-text">+&#x270D;</span>
        </button>
      </div>
    </div>
    <BLANKLINE>
    """
    # Local helper to cleanup options and acount for duplicate labels
    def update_options(opts):
        return update_choice_labels(
            [ FieldChoice(o) if is_string(o) else o 
              for o in opts
            ])

    # Local helper to render single option
    def select_option(opt):
        selected = ('' if opt.value != sel else ' selected="selected"')
        label    = (placeholder or "") if opt.value == "" else opt.choice_html()
        return '<option value="%s"%s>%s</option>'%(opt.value, selected, label)
    #
    select_template = (
        """<div class="row">\n"""+
        """  <div class="small-10 columns view-value less-new-button">\n"""+
        """    <select name="%(name)s">\n"""+
        """      %(options)s\n"""+
        """    </select>\n"""+
        """  </div>\n"""+
        """  <div class="small-2 columns view-value new-button left small-text-right">\n"""+
        """    <button type="submit" \n"""+
        """            name="%(name)s__new_edit" \n"""+
        """            value="New"\n"""+
        """            title="Define new or edit %(label)s">\n"""+
        """      <span class="select-edit-button-text">+&#x270D;</span>\n"""+
        """    </button>\n"""+
        """  </div>\n"""+
        """</div>\n"""+
        "")
    return select_template%(
        { 'name':       name
        , 'label':      label
        , 'options':    "\n      ".join(
            [ select_option(o) for o in update_options(opts) ])
        })

def render_choice_options(
        name, opts, sel, placeholder=None, select_class=None, escape_label=False):
    """
    Cf. `templates.field.annalist_edit_choice.html`.
    Like select, but without the "New" button.

    >>> print render_choice_options("foo", ["aa", "bb", "cc"], "bb")
    <select name="foo">
      <option value="aa">aa</option>
      <option value="bb" selected="selected">bb</option>
      <option value="cc">cc</option>
    </select>
    >>> print render_choice_options("foo", ["", "aa", "bb", "cc"], "", placeholder="(select)")
    <select name="foo">
      <option value="" selected="selected">(select)</option>
      <option value="aa">aa</option>
      <option value="bb">bb</option>
      <option value="cc">cc</option>
    </select>
    """
    # Local helper to render single option
    def select_option(opt):
        if is_string(opt):
            opt = FieldChoice(opt)
        selected = ('' if opt.value != sel else ' selected="selected"')
        label    = (placeholder or "") if opt.value == "" else opt.choice()
        if escape_label:
            label = escape(label)
        # label    = opt.label or opt.value or placeholder or ""
        return '<option value="%s"%s>%s</option>'%(opt.value, selected, label)
    #
    select_template = (
        """<select name="%(name)s"%(select_class)s>\n"""+
        """  %(options)s\n"""+
        """</select>"""+
        "")
    return select_template%(
        { 'name':           name
        , 'options':        "\n  ".join([ select_option(o) for o in opts ])
        , 'select_class':   ''' class="%s"'''%select_class if select_class is not None else ""
        })

#   -----------------------------------------------------------------------------
#
#   ----- Authorization support
#
#   -----------------------------------------------------------------------------

def create_user_permissions(parent, 
        user_id,
        user_permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG"]
        ):
    user_values = (
        { 'annal:type':             "annal:User"
        , 'rdfs:label':             "Test User"
        , 'rdfs:comment':           "User %s: permissions for %s in collection %s"%(user_id, "Test User", parent.get_id())
        , 'annal:user_uri':         "mailto:%s@%s"%(user_id, TestHost)
        , 'annal:user_permission':  user_permissions
        })
    user = AnnalistUser.create(parent, user_id, user_values)
    return user

def create_test_user(
        coll, user_id, user_pass="testpassword", 
        user_permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG"]
        ):
    django_user = User.objects.create_user(user_id, "%s@%s"%(user_id, TestHost), user_pass)
    django_user.first_name = "Test"
    django_user.last_name  = "User"
    django_user.save()
    if coll:
        user_perms = coll.create_user_permissions(
            user_id, "mailto:%s@%s"%(user_id, TestHost),
            "Test User",
            "User %s: permissions for %s in collection %s"%(user_id, "Test User", coll.get_id()),
            user_permissions)
    else:
        user_perms = None
    return (django_user, user_perms)

#   -----------------------------------------------------------------------------
#
#   ----- Context access utilities
#
#   -----------------------------------------------------------------------------

def context_field(field_desc):
    """
    Return a bound_field like object corresponding to suppliedfield details.
    The supplied parameter is a dictonary of field descriptiopn values, with a
    small number of additional keys recognized as setting values of the bound_field.
    """
    copy_field_desc  = field_desc.copy()
    entity_value     = copy_field_desc.pop('entity_value',     None)
    field_value      = copy_field_desc.pop('field_value',      None)
    field_edit_value = copy_field_desc.pop('field_edit_value', None)
    field_view_value = copy_field_desc.pop('field_view_value', None)
    field_options    = copy_field_desc.pop('field_options',    None)
    field = (
        { 'description':  copy_field_desc 
        })
    if field_value is not None:
        field['field_value']      = field_value
        field['field_edit_value'] = field_value
        field['field_view_value'] = field_value
    if field_edit_value is not None:
        field['field_edit_value'] = field_edit_value
    if field_view_value is not None:
        field['field_view_value'] = field_view_value
    if entity_value is not None:
        field['entity_value'] = entity_value
    if field_options is not None:
        field['options'] = field_options
    return field

def context_field_row(*fields):
    """
    Return a FieldDescription like object for a displayed row of fields
    in a view context.

    *fields      is a sequence of bound field like obects
    """
    field_row = context_field(
        { 'field_id':           "Row_fields"
        , 'field_name':         "Row_fields"
        , 'field_label':        "Fields in row"
        , 'field_value_mode':   "Value_direct"
        , 'field_render_type':  "FieldRow"
        , 'field_placement':    get_placement_classes('small:0,12')
        , 'row_field_descs':    fields
        })
    return field_row

def context_field_map(context):
    fields = context['fields']
    response = ["context_field_map: %d rows"%(len(fields))]
    for rn in range(len(fields)):
        row = fields[rn]
        if 'row_field_descs' in row:
            # Pick column from row
            cols = row['row_field_descs']
        else:
            # Field is not row-wrapped
            cols = [row]
        response.append("  row %d: %d cols"%(rn, len(cols)))
        for cn in range(len(cols)):
            field = cols[cn]
            response.append("    col %d, id %s, pos %s"%(cn, field['field_id'], field['field_placement']))
    return "\n".join(response)

def context_view_field(context, rownum, colnum):
    """
    Returns bound_field object corresponding to a given row&column in a view
    """
    row = context['fields'][rownum]
    if 'row_field_descs' in row.description:
        # Pick column from row
        field = row.description['row_field_descs'][colnum]
    else:
        # Field is not row-wrapped
        field = row
    if isinstance(field, FieldDescription):
        # fields in row are late-bound, so create a binding now
        entity_vals = context['fields'][rownum]['entity_value']
        extras      = context['fields'][rownum]['context_extra_values']
        field       = bound_field(field, entity_vals, context_extra_values=extras) 
    return field

def context_view_repeat_fields(context, repeat_bound_field):
    """
    Returns a list of lists of bound fields corresponding to 
    the items of a repeated value display field.

    Result is indexed by [row][field] to get individual bound fields
    """
    field_descs = repeat_bound_field.description['group_field_descs']
    extras      = repeat_bound_field['context_extra_values']
    bound_rows  = []
    for row_vals in repeat_bound_field.field_value:
        r = []
        for field_desc in field_descs:
            f = bound_field(field_desc, row_vals, context_extra_values=extras)
            r.append(f)
        bound_rows.append(r)
    return bound_rows

def context_bind_fields(context):
    """
    Fields in field rows are late bound so entity values do not appear in
    the actual context used.  This method binds entity values to fields
    so that all field values can be tested.
    """
    # bound_context = Context(context.flatten()) # Doesn't work for ContextList used for tests
    # Flatten values from context
    context_vals  = {}
    for k in context.keys():
        context_vals[k] = context[k]
    bound_context = Context(context_vals)
    bound_rows    = []
    for row in context['fields']:
        row_description = row['description']
        if 'row_field_descs' in row_description:
            entity_vals = row['entity_value']
            extras      = row['context_extra_values']
            bound_cols  = []
            for field in row_description['row_field_descs']:
                b_field = field
                if isinstance(field, FieldDescription):
                    # fields in row are late-bound, so create a binding now
                    b_field = bound_field(field, entity_vals, context_extra_values=extras) 
                bound_cols.append(b_field)
            bound_row = row.copy()
            bound_row._field_description['row_field_descs'] = bound_cols
            # print "@@@@ bound_row: "+repr(bound_row)
        else:
            bound_row = row
        bound_rows.append(bound_row)
    # print "@@@@ bound_rows[0]['description']['row_field_descs'][0]:\n"+repr(bound_rows[0]['description']['row_field_descs'][0])
    bound_context['fields'] = bound_rows
    return bound_context

def context_list_entities(context):
    """
    Returns list of entities to be displayed in list view
    """
    # log.info(context['List_rows'])
    if 'List_rows' in context:
        return context['List_rows']['field_value']
    elif 'entities' in context:
        return context['entities']
    log.warning("No entity list found in context %r"%(context.keys()))
    return None

def context_list_head_fields(context):
    """
    Returns field description used for accessing header information.
    """
    return context['fields']

def context_list_item_fields(context, entity):
    """
    Returns all list fields to be displayed as a list of bound_field values
    """
    # log.info(context['List_rows'])
    if 'List_rows' in context:
        fds = context['List_rows'].description['group_field_descs']
        return [ bound_field(fd, entity) for fd in fds ]
    elif 'fields' in entity:
        return entity['fields']
    log.warning("No field value found: context %r, entity %r"%(context.keys(), entity.keys()))
    return None

def context_list_item_field(context, entity, fid):
    """
    Returns indicated field to be displayed as a bound_field value
    """
    # log.info(context['List_rows'])
    if 'List_rows' in context:
        fd = context['List_rows'].description['group_field_descs'][fid]
        return bound_field(fd, entity)
    elif 'fields' in entity:
        return entity['fields'][fid]
    log.warning("No field value found: context %r, entity %r"%(context.keys(), entity.keys()))
    return None

def context_list_item_field_value(context, entity, fid):
    """
    Returns value of indicated field
    """
    return context_list_item_field(context, entity, fid)['field_value']

def check_context_field_value(test, context_field,
        field_value_type=None,
        field_value=None
        ):
    """
    Check field value in context field against supplied parameters.

    This function allows certain variations for robustness of tests.
    """
    context_field_value = context_field['field_value']
    if field_value is None:
        context_field_value = None
    elif field_value_type in ["annal:EntityRef", "annal:Type", "annal:View", "annal:List"]:
        context_field_value = extract_entity_id(context_field_value)
    if isinstance(field_value, (list, tuple)):
        if len(field_value) != len(context_field_value):
            #@@ Diagnostic:
            log.warning("@@ field_value:         %r"%(field_value,))
            log.warning("@@ context_field_value: %r"%(context_field_value,))
            #@@
        test.assertEqual(len(field_value), len(context_field_value))
        for i in range(len(field_value)):
            # print "@@ check_context_field_value (list): [%d] %r"%(i, field_value[i])
            # print "@@ check_context_field_value (list): [%d] %r (context)"%(i, context_field_value[i])
            log.debug("check_context_field_value (list): [%d] %r"%(i, field_value[i]))
            test.assertDictionaryMatch(context_field_value[i], field_value[i], prefix="[%d]"%i)
    elif isinstance(field_value, dict):
        test.assertDictionaryMatch(context_field_value, field_value[i], prefix="")
    else:
        test.assertEqual(context_field_value, field_value)
    return

def check_context_field(test, context_field,
        field_id=None,
        field_name=None,
        field_label=None,
        field_placeholder=None,
        field_property_uri=None,
        field_render_type=None,
        field_value_mode=None,
        field_value_type=None,
        field_placement=None,
        field_value=None,
        options=None,
        ):
    """
    Check values in context field against supplied parameters.

    This function allows certain variations for robustness of tests.
    """
    test.assertEqual(context_field.field_id,   field_id)
    test.assertEqual(context_field.field_name, field_name)
    if field_label:
        test.assertEqual(context_field.field_label, field_label)
    if field_placeholder:
        test.assertEqual(context_field.description['field_placeholder'], field_placeholder)
    if field_property_uri:
        test.assertEqual(context_field.description['field_property_uri'], field_property_uri)
    test.assertEqual(extract_entity_id(context_field.description['field_render_type']), field_render_type)
    test.assertEqual(extract_entity_id(context_field.description['field_value_mode']),  field_value_mode)
    test.assertEqual(context_field.description['field_value_type'],                     field_value_type)
    if options:
        if set(context_field['options']) != set(options):
            log.info("@@ Options expected: %r"%(context_field['options'],))
            log.info("@@ Options seen:     %r"%(options,))
            log.info("@@ context_field:    %r"%(context_field,))
        test.assertEqual(set(context_field['options']), set(options))
    if field_placement:
        test.assertEqual(context_field.description['field_placement'].field, field_placement)
    check_context_field_value(test, context_field, 
        field_value_type=field_value_type, 
        field_value=field_value
        )
    return

def check_context_list_field_value(test, context_field,
        field_value=None
        ):
    """
    Check field value according to type in context
    """
    check_context_field_value(test, context_field,
        field_value_type=context_field.description['field_value_type'],
        field_value=field_value
        )
    return

def check_field_list_context_fields(test, response, field_entities):
    """
    Check field list values in supplied context
    """
    head_fields = context_list_head_fields(response.context)
    test.assertEqual(len(head_fields), 1)       # One row of 4 cols..
    test.assertEqual(len(head_fields[0].description['row_field_descs']), 4)
    f0 = context_view_field(response.context, 0, 0)
    f1 = context_view_field(response.context, 0, 1)
    f2 = context_view_field(response.context, 0, 2)
    f3 = context_view_field(response.context, 0, 3)
    # 1st field
    check_context_field(test, f0,
        field_id=           "Entity_id",
        field_name=         "entity_id",
        field_label=        "Id",
        field_placeholder=  "(entity id)",
        field_property_uri= "annal:id",
        field_render_type=  "EntityId",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:EntityRef",
        field_placement=    "small-4 medium-3 columns"
        )
    # 2nd field
    check_context_field(test, f1,
        field_id=           "Field_render_type",
        field_name=         "Field_render_type",
        field_label=        "Render type",
        field_placeholder=  "(field render type)",
        field_property_uri= "annal:field_render_type",
        field_render_type=  "Enum_choice",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:EntityRef",
        field_placement=    "small-4 medium-3 columns"
        )
    # 3rd field
    check_context_field(test, f2,
        field_id=           "Field_value_type",
        field_name=         "Field_value_type",
        field_label=        "Value type",
        field_placeholder=  "(field value type)",
        field_property_uri= "annal:field_value_type",
        field_render_type=  "Identifier",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Identifier",
        field_placement=    "small-12 medium-3 columns show-for-medium-up"
        )
    # 4th field
    check_context_field(test, f3,
        field_id=           "Entity_label",
        field_name=         "Entity_label",
        field_label=        "Label",
        field_placeholder=  "(label)",
        field_property_uri= "rdfs:label",
        field_render_type=  "Text",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Text",
        field_placement=    "small-4 medium-3 columns"
        )
    # Selection of field entities from list
    entities = context_list_entities(response.context)
    entity_ids = [ context_list_item_field_value(response.context, e, 0) for e in entities ]
    for f in field_entities:
        for eid in range(len(entities)):
            item_fields = context_list_item_fields(response.context, entities[eid])
            if item_fields[0]['field_value'] == f[0]:
                # Expected field entity found in context data, check details...
                for fid in range(4):
                    item_field = item_fields[fid]
                    head_field = head_fields[0].description['row_field_descs'][fid]
                    for fkey in (
                            'field_id', 'field_name', 'field_label', 
                            'field_property_uri', 'field_render_type',
                            'field_placement', 'field_value_type'):
                        test.assertEqual(item_field.description[fkey], head_field[fkey])
                    # Check listed field values
                    check_context_list_field_value(test, item_field, f[fid])
                break
        else:
            test.fail("Field %s not found in context"%f[0])
    return

type_no_options   = [ FieldChoice('', label="(no options)") ]
def check_type_view_context_fields(test, response, 
        action="",
        entity_id="", orig_entity_id=None,
        type_id="_type",
        type_label="(?type_label)",
        type_comment="(?type_comment)",
        type_uri="(?type_uri)",
        type_supertype_uris="",
        type_view="Default_view", type_view_options=None,
        type_list="Default_list", type_list_options=None,
        type_aliases=[],
        type_data_ref=None,
        view_id="Type_view"
        ):
    # Common entity attributes
    # if orig_entity_id == "":
    #     orig_entity_id = entity_id
    test.assertEqual(response.context['entity_id'],        entity_id)
    test.assertEqual(response.context['orig_id'],          orig_entity_id)
    test.assertEqual(response.context['type_id'],          type_id)
    test.assertEqual(response.context['orig_type'],        type_id)
    test.assertEqual(response.context['coll_id'],          'testcoll')
    test.assertEqual(response.context['action'],           action)
    test.assertEqual(response.context['view_id'],          view_id)
    if type_data_ref:
        test.assertEqual(
            response.context['entity_data_ref'],      
            type_data_ref
            )
        test.assertEqual(
            response.context['entity_data_ref_json'], 
            type_data_ref+"?type=application/json"
            )
    # View fields
    test.assertEqual(len(response.context['fields']), 7)
    f0 = context_view_field(response.context, 0, 0)
    f1 = context_view_field(response.context, 1, 0)
    f2 = context_view_field(response.context, 2, 0)
    f3 = context_view_field(response.context, 3, 0)
    f4 = context_view_field(response.context, 4, 0)
    f5 = context_view_field(response.context, 5, 0)
    f6 = context_view_field(response.context, 5, 1)
    f7 = context_view_field(response.context, 6, 0)
    # 1st field - Id
    check_context_field(test, f0,
        field_id=           "Type_id",
        field_name=         "entity_id",
        field_label=        "Type Id",
        field_placeholder=  "(type id)",
        field_property_uri= "annal:id",
        field_render_type=  "EntityId",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:EntityRef",
        field_placement=    "small-12 medium-6 columns",
        field_value=        entity_id,
        options=            type_no_options
        )
    # 2nd field - Label
    check_context_field(test, f1,
        field_id=           "Type_label",
        field_name=         "Type_label",
        field_label=        "Label",
        field_placeholder=  "(label)",
        field_property_uri= "rdfs:label",
        field_render_type=  "Text",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Text",
        field_placement=    "small-12 columns",
        field_value=        type_label,
        options=            type_no_options
        )
    # 3rd field - comment
    type_comment_placeholder = (
        "(type description)"
        )
    check_context_field(test, f2,
        field_id=           "Type_comment",
        field_name=         "Type_comment",
        field_label=        "Comment",
        field_placeholder=  type_comment_placeholder,
        field_property_uri= "rdfs:comment",
        field_render_type=  "Markdown",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Richtext",
        field_placement=    "small-12 columns",
        field_value=        type_comment,
        options=            type_no_options
        )
    # 4th field - URI
    type_uri_placeholder = (
        "(Type URI or CURIE)"
        )
    check_context_field(test, f3,
        field_id=           "Type_uri",
        field_name=         "Type_uri",
        field_label=        "Type URI",
        field_placeholder=  type_uri_placeholder,
        field_property_uri= "annal:uri",
        field_render_type=  "Identifier",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Identifier",
        field_value=        type_uri,
        options=            type_no_options
        )
    # 5th field - Supertype URIs
    type_supertype_uris_placeholder = (
        "(Supertype URIs or CURIEs)"
        )
    check_context_field(test, f4,
        field_id=           "Type_supertype_uris",
        field_name=         "Type_supertype_uris",
        field_label=        "Supertype URIs",
        field_placeholder=  type_supertype_uris_placeholder,
        field_property_uri= "annal:supertype_uri",
        field_render_type=  "Group_Set_Row",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Type_supertype_uri",
        field_value=        type_supertype_uris,
        options=            type_no_options
        )
    # 6th field - view id
    type_view_id_placeholder = (
        "(view id)"
        )
    check_context_field(test, f5,
        field_id=           "Type_view",
        field_name=         "Type_view",
        field_label=        "Default view",
        field_placeholder=  type_view_id_placeholder,
        field_property_uri= "annal:type_view",
        field_render_type=  "Enum_optional",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:View",
        field_placement=    "small-12 medium-6 columns",
        field_value=        type_view,
        options=            type_view_options
        )
    # 7th field - list id
    type_list_id_placeholder = (
        "(list id)"
        )
    check_context_field(test, f6,
        field_id=           "Type_list",
        field_name=         "Type_list",
        field_label=        "Default list",
        field_placeholder=  type_list_id_placeholder,
        field_property_uri= "annal:type_list",
        field_render_type=  "Enum_optional",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:List",
        field_placement=    "small-12 medium-6 columns",
        field_value=        type_list,
        options=            type_list_options
        )
    # 8th field - field aliases
    type_aliases_placeholder = (
        "(field aliases)"
        )
    check_context_field(test, f7,
        field_id=           "Type_aliases",
        field_name=         "Type_aliases",
        field_label=        "Field aliases",
        field_placeholder=  type_aliases_placeholder,
        field_property_uri= "annal:field_aliases",
        field_render_type=  "Group_Seq_Row",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Type_alias",
        field_placement=    "small-12 columns",
        field_value=        type_aliases,
        options=            type_no_options
        )
    return

entity_no_options   = [ FieldChoice('', label="(no options)") ]
def check_default_view_context_fields(test, response, 
        action="",
        entity_id="", orig_entity_id=None,
        type_id="testtype", type_options=None,
        entity_label="(?entity_label)",
        entity_comment="(?entity_comment)",
        entity_data_ref=None,
        view_id="Type_view"
        ):
    # Common entity attributes
    # if orig_entity_id == "":
    #     orig_entity_id = entity_id
    test.assertEqual(response.context['entity_id'],        entity_id)
    test.assertEqual(response.context['orig_id'],          orig_entity_id)
    test.assertEqual(response.context['type_id'],          type_id)
    test.assertEqual(response.context['orig_type'],        type_id)
    test.assertEqual(response.context['coll_id'],          'testcoll')
    test.assertEqual(response.context['action'],           action)
    test.assertEqual(response.context['view_id'],          view_id)
    if entity_data_ref:
        test.assertEqual(
            response.context['entity_data_ref'],      
            entity_data_ref
            )
        test.assertEqual(
            response.context['entity_data_ref_json'], 
            entity_data_ref+"?type=application/json"
            )
    # View fields
    # log.info("@@@@ response.context['fields'] %r"%(response.context['fields'],))
    test.assertEqual(len(response.context['fields']), 3)
    f0 = context_view_field(response.context, 0, 0)
    f1 = context_view_field(response.context, 0, 1)
    f2 = context_view_field(response.context, 1, 0)
    f3 = context_view_field(response.context, 2, 0)
    # 1st field - Id
    check_context_field(test, f0,
        field_id=           "Entity_id",
        field_name=         "entity_id",
        field_label=        "Id",
        field_placeholder=  "(entity id)",
        field_property_uri= "annal:id",
        field_render_type=  "EntityId",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:EntityRef",
        field_placement=    "small-12 medium-6 columns",
        field_value=        entity_id,
        options=            entity_no_options
        )
    # 2nd field - type id
    check_context_field(test, f1,
        field_id=           "Entity_type",
        field_name=         "entity_type",
        field_label=        "Type",
        field_placeholder=  "(type id)",
        field_property_uri= "annal:type_id",
        field_render_type=  "EntityTypeId",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:EntityRef",
        field_placement=    "small-12 medium-6 columns",
        field_value=        type_id,
        options=            type_options
        )
    # 3rd field - Label
    check_context_field(test, f2,
        field_id=           "Entity_label",
        field_name=         "Entity_label",
        field_label=        "Label",
        field_placeholder=  "(label)",
        field_property_uri= "rdfs:label",
        field_render_type=  "Text",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Text",
        field_placement=    "small-12 columns",
        field_value=        entity_label,
        options=            entity_no_options
        )
    # 4th field - comment
    check_context_field(test, f3,
        field_id=           "Entity_comment",
        field_name=         "Entity_comment",
        field_label=        "Comment",
        field_placeholder=  "(description)",
        field_property_uri= "rdfs:comment",
        field_render_type=  "Markdown",
        field_value_mode=   "Value_direct",
        field_value_type=   "annal:Richtext",
        field_placement=    "small-12 columns",
        field_value=        entity_comment,
        options=            entity_no_options
        )
    return

def check_field_record(test, field_record,
        field_id=None,
        field_ref=None,
        field_types=None,
        field_name=None,
        field_type_id=None,
        field_type=None,
        field_uri=None,
        field_url=None,
        field_label=None,
        field_help=None,
        field_render_type=None,
        field_value_mode=None,
        field_property_uri=None,
        field_placement=None,
        field_entity_type=None,
        field_value_type=None,
        field_placeholder=None,
        field_tooltip=None,
        field_default=None,
        ):
        if field_id:
            test.assertEqual(field_id,           field_record['annal:id'])
        if field_ref:
            test.assertEqual(field_ref,          field_record['@id'])
        if field_types:
            test.assertEqual(field_types,        field_record['@type'])
        if field_type_id:
            test.assertEqual(field_type_id,      field_record['annal:type_id'])
        if field_type:
            test.assertEqual(field_type,         field_record['annal:type'])
        if field_uri:
            test.assertEqual(field_uri,          field_record['annal:uri'])
        if field_url:
            test.assertEqual(field_url,          field_record['annal:url'])
        if field_label:
            test.assertEqual(field_label,        field_record['rdfs:label'])
        if field_help:
            test.assertEqual(field_help,         field_record['rdfs:comment'])

        if field_name:
            test.assertEqual(field_name,         field_record['annal:field_name'])
        if field_render_type:
            test.assertEqual(
                field_render_type,  
                extract_entity_id(field_record['annal:field_render_type'])
                )
        if field_value_mode:
            test.assertEqual(
                field_value_mode,   
                extract_entity_id(field_record['annal:field_value_mode'])
                )
        if field_property_uri:
            test.assertEqual(field_property_uri, field_record['annal:property_uri'])
        if field_placement:
            test.assertEqual(field_placement,    field_record['annal:field_placement'])
        if field_entity_type:
            test.assertEqual(field_entity_type,  field_record['annal:field_entity_type'])
        if field_value_type:
            test.assertEqual(field_value_type,   field_record['annal:field_value_type'])
        if field_placeholder:
            test.assertEqual(field_placeholder,  field_record['annal:placeholder'])
        if field_tooltip:
            test.assertStarts(field_tooltip,     field_record['annal:tooltip'])
        if field_default:
            test.assertEqual(field_default,      field_record['annal:default_value'])
        return

#   -----------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
annalist_manager tests
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

def get_source_root():
	import os
	here     = os.path.dirname(__file__)
	src_root = os.path.normpath(os.path.join(here, "../../"))
	return src_root

# End.



"""
Test module for annalist-manager collection data management commands
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os

from utils.py3porting    import StringIO
from utils.StdoutContext import SwitchStdout, SwitchStderr

import annalist
from annalist.util              import replacetree, removetree

from annalist_manager.tests     import get_source_root
from annalist_manager.tests     import test_annalist_base
from annalist_manager.am_main   import runCommand

#   -----------------------------------------------------------------------------
#
#   Tests
#
#   -----------------------------------------------------------------------------

class AnnalistManagerCollTest(test_annalist_base.AnnalistManagerTestBase):

    @classmethod
    def setUpTestData(cls):
        cls.setup_annalist_manager_test()
        cls.init_site_database()
        cls.create_site_data()
        return

    def setUp(self):
        coll_dir = self.colldir("Resource_defs")
        if os.path.isdir(coll_dir):
            removetree(coll_dir)
        return

    def tearDown(self):
        return

    #   -----------------------------------------------------------------------------
    #   Tests
    #   -----------------------------------------------------------------------------

    def test_installcollection(self):
        coll_id = "Resource_defs"
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                [ "annalist-manager", "installcollection"
                , coll_id
                , "--config=runtests"
                ])
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(
            stdoutlines[0], 
            "Installing collection '%s' from data directory '%s'"%
              (coll_id, self.collsrc(coll_id))
            )
        collexists = os.path.isdir(self.colldir(coll_id))
        self.assertTrue(collexists, "%s collection created?"%(coll_id))
        return

    def test_copycollection(self):
        coll_id1 = "Resource_defs"
        coll_id2 = "Resource_defs_copy"
        # Create source
        self.installcoll(coll_id1)
        collexists = os.path.isdir(self.colldir(coll_id1))
        self.assertTrue(collexists, "%s created?"%coll_id1)
        # Now copy
        coll_dir = self.colldir(coll_id2)
        if os.path.isdir(coll_dir):
            removetree(coll_id2)
        collexists = os.path.isdir(self.colldir(coll_id2))
        self.assertFalse(collexists, "%s absent?"%coll_id2)
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                [ "annalist-manager", "copycollection"
                , coll_id1, coll_id2
                , "--config=runtests"
                ])
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(
            stdoutlines[0], 
            "Copying collection '%s' to '%s'"%(coll_id1, coll_id2)
            )
        collexists = os.path.isdir(self.colldir(coll_id2))
        self.assertTrue(os.path.isdir(self.colldir(coll_id1)), "%s created?"%coll_id2)
        return

    def test_migrationreport(self):
        coll_id1 = "Resource_defs"
        coll_id2 = "Resource_defs_copy"
        # Create sources
        self.installcoll(coll_id1)
        collexists = os.path.isdir(self.colldir(coll_id1))
        self.assertTrue(collexists, "%s created?"%coll_id1)
        self.copycoll(coll_id1, coll_id2)
        collexists = os.path.isdir(self.colldir(coll_id2))
        self.assertTrue(collexists, "%s created?"%coll_id2)
        # Now generate migration report
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                [ "annalist-manager", "migrationreport"
                , coll_id1, coll_id2
                , "--config=runtests"
                ])
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(
            stdoutlines[0], 
            "# Migration report from collection '%s' to '%s' #"%(coll_id1, coll_id2)
            )
        return

    def test_migratecollection(self):
        coll_id = "Resource_defs"
        # Create source
        self.installcoll(coll_id)
        collexists = os.path.isdir(self.colldir(coll_id))
        self.assertTrue(collexists, "%s created?"%coll_id)
        # Now migrate
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                [ "annalist-manager", "migratecollection"
                , coll_id
                , "--config=runtests"
                ])
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(
            stdoutlines[0], 
            "Apply data migrations in collection '%s'"%(coll_id,)
            )
        return

    def test_migrateallcollections(self):
        coll_id1 = "Resource_defs"
        coll_id2 = "Resource_defs_copy"
        coll_id3 = "Concept_defs"
        coll_id4 = "Journal_defs"
        # Create sources
        self.installcoll(coll_id1)
        collexists = os.path.isdir(self.colldir(coll_id1))
        self.assertTrue(collexists, "%s created?"%coll_id1)
        self.copycoll(coll_id1, coll_id2)
        collexists = os.path.isdir(self.colldir(coll_id2))
        self.assertTrue(collexists, "%s created?"%coll_id2)
        self.installcoll(coll_id3)
        collexists = os.path.isdir(self.colldir(coll_id3))
        self.assertTrue(collexists, "%s created?"%coll_id3)
        self.installcoll(coll_id4)
        collexists = os.path.isdir(self.colldir(coll_id4))
        self.assertTrue(collexists, "%s created?"%coll_id4)
        # Now migrate
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "migrateallcollections"
                    , "--config=runtests"
                    ])
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        expectlines = (
            [ "---- Processing '%s'"%(cid) 
              for cid in (coll_id1, coll_id2, coll_id3, coll_id4)
            ])
        expectlines.append("")
        self.assertEqual(
            stdoutlines[0], 
            "Apply data migrations in all collections:"
            )
        self.assertIn(stdoutlines[1], expectlines)
        self.assertIn(stdoutlines[2], expectlines)
        self.assertIn(stdoutlines[3], expectlines)
        self.assertIn(stdoutlines[4], expectlines)
        self.assertEqual(
            stdoutlines[5], 
            "Data migrations complete."
            )
        return

# End.



"""
Annalist generic view definition
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import os.path
import json
import markdown
import traceback

from django.http                    import HttpResponse
from django.http                    import HttpResponseRedirect
from django.template                import loader
from django.views                   import generic
from django.views.decorators.csrf   import csrf_exempt
from django.core.urlresolvers       import resolve, reverse
from django.conf                    import settings

import login.login_views

from utils.py3porting               import urljoin
from utils.ContentNegotiationView   import ContentNegotiationView

import annalist
from annalist                       import message
from annalist                       import layout
from annalist                       import util
from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist.models.site           import Site
from annalist.models.annalistuser   import (
    AnnalistUser, 
    site_default_user_id, site_default_user_uri, 
    default_user_id, default_user_uri, 
    unknown_user_id, unknown_user_uri
    )

from annalist.views.uri_builder     import uri_with_params, continuation_params, uri_params

#   -------------------------------------------------------------------------------------------
#
#   Utility methods and data
#
#   -------------------------------------------------------------------------------------------

LOGIN_URIS = None   # Populated by first call of `authenticate`

#   -------------------------------------------------------------------------------------------
#
#   Generic Annalist view (contains logic applicable to all pages)
#
#   -------------------------------------------------------------------------------------------

class AnnalistGenericView(ContentNegotiationView):
    """
    Common base class for Annalist views
    """

    def __init__(self):
        super(AnnalistGenericView, self).__init__()
        self._sitebaseuri      = reverse("AnnalistHomeView")
        # self._sitebasedir      = os.path.join(settings.BASE_DATA_DIR, layout.SITE_DIR)
        self._sitebasedir      = settings.BASE_SITE_DIR
        self._site             = None
        self._site_data        = None
        self._user_perms       = {}
        return

    # @@TODO: make host parameter required in the following?

    def site(self, host=""):
        if not self._site:
            self._site = Site(self._sitebaseuri, self._sitebasedir, host=host)
        return self._site

    def site_data(self, host=""):
        if not self._site_data:
            self._site_data = self.site(host=host).site_data()
        if not self._site_data:
            log.error("views.generic.site_data: failed to load site data (%s)"%
                      self.site(host=host)._dir_path_uri()[1])
        return self._site_data

    def error(self, values, continuation_url=None):
        """
        Construct HTTP error response.

        This is an application-specific override of a method defined 
        in ContentNegotiationView.

        values      is a dictionary of values to be passed as a context to the
                    error display page.  Typically, this contains more details
                    about the error (cf. ContentNegotiationView.errorvalues)
        continutation_url
                    is a URL to (re)display when any error is dismissed or has 
                    otherwise been handled.
        """
        # log.info(
        #     "AnnalistGenericView.error: values %r, continuation_url %s"%(values, continuation_url)
        #     )
        template = loader.get_template('annalist_error.html')
        context = dict(values)
        if continuation_url:
            context['continuation_url']   = continuation_url
            context['continuation_param'] = uri_params({ 'continuation_url': continuation_url })
        return HttpResponse(template.render(context), status=values['status'], reason=values['reason'])

    def view_uri(self, viewname, **kwargs):
        """
        Return view URI given view name and any additional arguments
        """
        return reverse(viewname, kwargs=kwargs)

    def get_collection_view_url(self, coll_id):
        """
        Return view (root) URL for specified collection
        """
        return self.view_uri(
            "AnnalistCollectionView", 
            coll_id=coll_id
            )

    def get_site_base_url(self):
        """
        Return base URL for current site
        """
        return self.view_uri("AnnalistHomeView")

    def get_collection_base_url(self, coll_id):
        """
        Return base URL for specified collection
        """
        return urljoin(self.get_collection_view_url(coll_id), layout.COLL_BASE_REF)

    def get_entity_base_url(self, coll_id, type_id, entity_id):
        """
        Return base URL for specified entity
        """
        return self.view_uri(
            "AnnalistEntityAccessView", 
            coll_id=coll_id, type_id=type_id, entity_id=entity_id
            )

    def get_list_base_url(self, coll_id, type_id, list_id):
        """
        Return base URL for specified list, which is one of:

            .../coll_id/d/
            .../coll_id/d/type_id/
            .../coll_id/l/list_id/
            .../coll_id/l/list_id/type_id/
        """
        if list_id is None:
            if type_id is None:
                list_url = self.view_uri(
                    "AnnalistEntityDefaultListAll", 
                    coll_id=coll_id
                    )
            else:
                list_url = self.view_uri(
                    "AnnalistEntityDefaultListType", 
                    coll_id=coll_id, type_id=type_id
                    )
        else:
            list_url = self.view_uri(
                "AnnalistEntityGenericList", 
                coll_id=coll_id, list_id=list_id, type_id=type_id
                )
        return list_url

    def resource_response(self, resource_file, resource_type, links={}):
        """
        Construct response containing body of referenced resource (or list),
        with supplied resource_type as its content_type
        """
        # @@TODO: assumes response can reasonably be held in memory;
        #         consider 'StreamingHttpResponse'?
        response = HttpResponse(content_type=resource_type)
        response = self.add_link_header(response, links)
        response.write(resource_file.read())
        return response

    def continuation_next(self, request_dict={}, default_cont=None):
        """
        Returns a continuation URL to be used when returning from the current view,
        or the supplied default if no continuation is specified for the current view.
        """
        return request_dict.get("continuation_url") or default_cont or None

    def continuation_here(self, request_dict={}, default_cont="", base_here=None):
        """
        Returns a URL that returns control to the current page, to be passed as a
        continuation_uri parameter to any subsidiary pages invoked.  Such continuation 
        URIs are cascaded, so that the return URI includes a the `continuation_url` 
        for the current page.

        request_dict    is a request dictionary that is expected to contain the 
                        continuation_url value to use, and other parameters to 
                        be included an any continuation back to the current page.
        default_cont    is a default continuation URI to be used for returning from 
                        the current page if the current POST request does not specify
                        a continuation_url query parameter.
        base_here       if specified, overrides the current request path as the base URI
                        to be used to return to the currently displayed page (e.g. when
                        current request URI is non-idempotent, such as creating a new 
                        entity).
        """
        # Note: use default if request/form parameter is present but blank:
        if not base_here:
            base_here = self.get_request_path()
        continuation_next = self.continuation_next(request_dict, default_cont)
        return uri_with_params(base_here, 
            continuation_params({"continuation_url": continuation_next}, request_dict)
            )

    def info_params(self, info_message, info_head=message.ACTION_COMPLETED):
        """
        Returns a URI query parameter dictionary with details that are used to generate an
        information message.
        """
        return {"info_head": info_head, "info_message": info_message}

    def redirect_info(self, viewuri, view_params={}, info_message=None, info_head=message.ACTION_COMPLETED):
        """
        Redirect to a specified view with an information/confirmation message for display

        (see templates/base_generic.html for display details)
        """
        redirect_uri = uri_with_params(viewuri, view_params, self.info_params(info_message, info_head))
        return HttpResponseRedirect(redirect_uri)

    def error_params(self, error_message, error_head=message.INPUT_ERROR):
        """
        Returns a URI query parameter string with details that are used to generate an
        error message.
        """
        return {"error_head": error_head, "error_message": error_message}

    def redirect_error(self, 
        viewuri, view_params={}, error_message=None, error_head=message.INPUT_ERROR
        ):
        """
        Redirect to a specified view with an error message for display

        (see templates/base_generic.html for display details)
        """
        redirect_uri = uri_with_params(
            viewuri, view_params, self.error_params(error_message, error_head=error_head)
            )
        return HttpResponseRedirect(redirect_uri)

    def check_site_data(self):
        """
        Check that site data is present and accessible.  If not, return an HTTP error
        response, otherwise None.
        """
        site_data = self.site_data()
        if not site_data:
            return self.error(
                self.errorvalues(500, "Internal server error", 
                    "Resource %(request_uri)s: Unable to load Annalist site data"
                    )
                )
        return None

    # Authentication and authorization
    def authenticate(self, continuation_url):
        """
        Return None if required authentication is present, otherwise
        an appropriate login redirection response.

        continuation_url    is a URL that to be retrieved and processed 
                            when the authentication process completes
                            (or is aborted).

        self.credential is set to credential that can be used to access resource
        """
        # Cache copy of URIs to use with login
        global LOGIN_URIS
        if LOGIN_URIS is None:
            LOGIN_URIS = (
                { "login_form_url":     self.view_uri('LoginUserView')
                , "login_post_url":     self.view_uri('LoginPostView')
                , "login_done_url":     self.view_uri('OIDC_AuthDoneView')
                , "user_profile_url":   self.view_uri('AnnalistProfileView')
                })
        return login.login_views.confirm_authentication(self, 
            continuation_url=continuation_url,
            **LOGIN_URIS
            )

    def get_user_identity(self):
        """
        returns the username and authenticated URI for the user making the 
        current request, as a pair (username, URI)
        """
        user = self.request.user
        if user.is_authenticated:
            return (user.username, "mailto:"+user.email)
        return ("_unknown_user_perms", "annal:User/_unknown_user_perms")

    def get_user_permissions(self, collection, user_id, user_uri):
        """
        Get a user permissions record (AnnalistUser).

        To return a value, both the user_id and the user_uri (typically a mailto: URI, but
        may be any *authenticated* identifier) must match.  This is to prevent access to 
        records of a deleted account being granted to a new account created with the 
        same user_id (username).

        This function includes any collection- or site- default permissions in the 
        set of permissions returned.

        Permissions are cached in the view object so that the permissions record is read at 
        most once for any HTTP request. 

        collection      the collection for which permissions are required.
        user_id         local identifier for the user permnissions to retrieve.
        user_uri        authenticated identifier associated with the user_id.  That is,
                        the authentication service used is presumed to confirm that
                        the identifier belongs to the user currently logged in with
                        the supplied username.

        returns an AnnalistUser object containing permissions for the identified user.
        """
        coll_id = collection.get_id() if collection else layout.SITEDATA_ID
        if coll_id not in self._user_perms:
            # if collection:
            #     log.info("user_id %s (%s), coll_id %s, %r"%
            #         (user_id, user_uri, collection.get_id(), collection.get_user_permissions(user_id, user_uri))
            #         )
            parentcoll = collection or self.site()
            if user_id == unknown_user_id:
                # Don't apply collection default-user permissions if no authenticated user
                user_perms = parentcoll.get_user_permissions(unknown_user_id, unknown_user_uri)
            else:
                # Combine user permissions with default-user permissions for collection
                default_perms = parentcoll.get_user_permissions(site_default_user_id, default_user_uri)
                if not default_perms:
                    default_perms = parentcoll.get_user_permissions(default_user_id, default_user_uri)
                user_perms    = parentcoll.get_user_permissions(user_id, user_uri) or default_perms
                user_perms[ANNAL.CURIE.user_permission] = list(
                    set(user_perms[ANNAL.CURIE.user_permission]) | 
                    set(default_perms[ANNAL.CURIE.user_permission])
                    ) 
            self._user_perms[coll_id] = user_perms
            # log.debug("get_user_permissions %r"%(self._user_perms,))
        return self._user_perms[coll_id]

    def get_permissions(self, collection):
        """
        Get permissions for the current user
        """
        user_id, user_uri = self.get_user_identity()
        return self.get_user_permissions(collection, user_id, user_uri)

    def get_message_data(self):
        """
        Returns a dictionary of message data that can be passed inthe request parameters
        to be displayed on a different page.
        """
        messagedata = {}
        def uri_param_val(msg_name, hdr_name, hdr_default):
            """
            Incorporate values from the incoming URI into the message data.
            """
            message = self.request.GET.get(msg_name, None)
            if message:
                messagedata[msg_name] = message
                messagedata[hdr_name] = self.request.GET.get(hdr_name, hdr_default)
            return
        uri_param_val("info_message",  "info_head",  message.ACTION_COMPLETED)
        uri_param_val("error_message", "error_head", message.INPUT_ERROR) 
        return messagedata

    def authorize(self, scope, collection, continuation_url=None):
        """
        Return None if user is authorized to perform the requested operation,
        otherwise appropriate 401 Authorization Required or 403 Forbidden response.
        May be called with or without an authenticated user.

        scope       indication of the operation  requested to be performed.
                    e.g. "VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG", ...
        collection  is the collection to which the requested action is directed,
                    or None if the test is against site-level permissions.
        continutation_url
                    is a URL to (re)display when any error is dismissed or has 
                    otherwise been handled.
        """
        user_id, user_uri = self.get_user_identity()
        coll_id = collection.get_id() if collection else "(site)"
        if not util.valid_id(user_id):
            log.warning("Invalid user_id %s, URI %s"%(user_id, user_uri))
            message="Bad request to %(request_uri)s: invalid user_id: '"+user_id+"'"
            return self.error(self.error400values(message=message), continuation_url=continuation_url)
        user_perms = self.get_user_permissions(collection, user_id, user_uri)
        if not user_perms:
            log.warning("No user permissions found for user_id %s, URI %s"%(user_id, user_uri))
            log.warning("".join(traceback.format_stack()))
            return self.error(self.error403values(scope=scope), continuation_url=continuation_url)
        # log.info("Authorize %s in %s, %s, %r"%(user_id, coll_id, scope, user_perms[ANNAL.CURIE.user_permission]))
        # user_perms is an AnnalistUser object
        coll_id = collection.get_id() if collection else "(No coll)"
        if scope not in user_perms[ANNAL.CURIE.user_permission]:
            if user_id == unknown_user_id:
                err = self.error401values(scope=scope)
            else:
                err = self.error403values(scope=scope)
            return self.error(err, continuation_url=continuation_url)
        return None

    def form_action_auth(self, action, auth_collection, perm_required, continuation_url=None):
        """
        Check that the requested form action is authorized for the current user.

        action          is the requested action: new, edit, copy, etc.
        auth_collection is the collection to which the requested action is directed,
                        or None if the test is against site-level permissions
                        (which should be stricter than all collections).
        perm_required   is a data dependent dictionary that maps from action to
                        permissions required to perform the action.  The structure
                        is similar to that of 'action_scope' (below) that provides
                        a fallback mapping.
        continutation_url
                        is a URL to (re)display when any error is dismissed or has 
                        otherwise been handled.

        Returns None if the desired action is authorized for the current user, otherwise
        an HTTP response value to return an error condition.
        """
        if action in perm_required:
            auth_scope = perm_required[action]
        else:
            log.warning("form_action_auth: unknown action: %s"%(action))
            log.warning("perm_required: %r"%(perm_required,))
            auth_scope = "UNKNOWN"
        return self.authorize(auth_scope, auth_collection, continuation_url=continuation_url)

    # Entity access

    def get_entity(self, entity_id, typeinfo, action):
        """
        Create local entity object or load values from existing.

        entity_id       entity id to create or load
        typeinfo        EntityTypeInfo object for the entity
        action          is the requested action: new, edit, copy, view

        Returns an object of the appropriate type.

        If an existing entity is accessed, values are read from storage, 

        If the identified entity does not exist and `action` is "new" then 
        a new entity is initialized (but not saved), otherwise the entity 
        value returned is None.
        """
        # log.info(
        #     "AnnalistGenericView.get_entity id %s, parent %s, action %s, altparent %s"%
        #     (entity_id, typeinfo.entityparent, action, typeinfo.entityaltparent)
        #     )
        entity = typeinfo.get_entity(entity_id, action)
        if entity is None:
            # Log diagnostics for missing entity
            parent_id    = typeinfo.entityparent.get_id()
            altparent_id = (
                typeinfo.entityaltparent.get_id() if typeinfo.entityaltparent 
                else "(none)"
                )
            # log.info(
            #     "AnnalistGenericView.get_entity id %s, parent %s, action %s, altparent %s"%
            #     (entity_id, typeinfo.entityparent, action, typeinfo.entityaltparent)
            #     )
            log.info(
                "AnnalistGenericView.get_entity, not found: parent_id %s, altparent_id %s, entity_id %s"%
                (parent_id, altparent_id, entity_id)
                )
        return entity

    # HTML rendering

    @ContentNegotiationView.accept_types(["text/html", "application/html", "*/*"])
    def render_html(self, resultdata, template_name, links=[]):
        """
        Construct an HTML response based on supplied data and template name.

        Also contains logic to interpolate message values from the incoming URI,
        for error and confirmation message displays.  These additional message
        displays are commonly handled by the "base_generic.html" underlay template.
        """
        def uri_param_val(name, default):
            """
            Incorporate values from the incoming URI into the result data, if not already defined.
            """
            if name not in resultdata:
                resultdata[name] = self.request.GET.get(name, default)
            return
        uri_param_val("info_head",       message.ACTION_COMPLETED)
        uri_param_val("info_message",    None)
        uri_param_val("error_head",      message.INPUT_ERROR) 
        uri_param_val("error_message",   None)
        resultdata["annalist_version"] = annalist.__version__
        if 'help_filename' in resultdata:
            help_filepath = os.path.join(
                settings.SITE_SRC_ROOT, 
                "annalist/views/help/%s.html"%resultdata['help_filename']
                )
            if os.path.isfile(help_filepath):
                with open(help_filepath, "r") as helpfile:
                    resultdata['help_text'] = helpfile.read()
        if 'help_markdown' in resultdata:
            resultdata['help_text'] = markdown.markdown(resultdata['help_markdown'])
        template = loader.get_template(template_name)
        context  = resultdata
        # log.debug("render_html - data: %r"%(resultdata))
        response = HttpResponse(template.render(context, request=self.request))
        if "entity_data_ref" in resultdata:
            alt_link = [ { "ref": resultdata["entity_data_ref"], "rel": "alternate" } ]
        else:
            alt_link = []
        response = self.add_link_header(response, links=alt_link+links)
        return response

    # JSON and Turtle content negotiation and redirection

    @ContentNegotiationView.accept_types(["application/json", "application/ld+json"])
    def redirect_json(self, jsonref, links=[]):
        """
        Construct a redirect response to access JSON data at the designated URL.

        jsonref     is the URL from which JSON data may be retrieved.
        links       is an optional array of link values to be added to the HTTP response
                    (see method add_link_header for description).

        Returns an HTTP redirect response object if the current request is for JSON data,
        otherwise None.
        """
        response = HttpResponseRedirect(jsonref)
        response = self.add_link_header(response, links=links)
        return response

    @ContentNegotiationView.accept_types(["text/turtle", "application/x-turtle", "text/n3"])
    def redirect_turtle(self, turtleref, links=[]):
        """
        Construct a redirect response to access Turtle data at the designated URL.

        turtleref   is the URL from which Turtle data may be retrieved.
        links       is an optional array of link values to be added to the HTTP response
                    (see method add_link_header for description).

        Returns an HTTP redirect response object if the current request is for Turtle data,
        otherwise None.
        """
        response = HttpResponseRedirect(turtleref)
        response = self.add_link_header(response, links=links)
        return response

    def add_link_header(self, response, links=[]):
        """
        Add HTTP link header to response, and return the updated response.

        response    response to be returned
        links       is an optional array of link valuyes to be added to the HTTP response, 
                    where each link is specified as:
                      { "rel": <relation-type>
                      , "ref": <target-url>
                      }
        """
        link_headers = []
        for l in links:
            link_headers.append('''<%(ref)s>; rel="%(rel)s"'''%l)
        if len(link_headers) > 0:
            response["Link"] = ",".join(link_headers)
        return response

    # Default view methods return 405 Forbidden

    def get(self, request, *args, **kwargs):
        return self.error(self.error405values())

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.error(self.error405values())

    def post(self, request, *args, **kwargs):
        return self.error(self.error405values())

    def delete(self, request, *args, **kwargs):
        return self.error(self.error405values())


# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for views and edits that use field subproperty URIs.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.contrib.auth.models     import User
from django.test                    import TestCase
from django.test.client             import Client

from annalist                       import layout
from annalist.identifiers           import ANNAL, RDFS

from annalist.models.entitytypeinfo import EntityTypeInfo
from annalist.models.entity         import EntityRoot, Entity
from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.entitydata     import EntityData

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site,
    init_annalist_test_coll,
    init_annalist_named_test_coll,
    resetSitedata
    )
from .entity_testutils import (
    collection_create_values,
    create_test_user, create_user_permissions,
    context_view_field, context_view_repeat_fields,
    context_list_entities,
    context_list_head_fields, context_list_item_fields,
    context_list_item_field, context_list_item_field_value
    )
from .entity_testentitydata import (
    entity_url, entitydata_edit_url, 
    entitydata_list_type_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values, 
    )
from .entity_testtypedata import (
    recordtype_create_values, 
    )

#   -----------------------------------------------------------------------------
#
#   Field alias tests
#
#   -----------------------------------------------------------------------------

class FieldSubpropertyTest(AnnalistTestCase):
    """
    Tests for field alias values
    """

    def setUp(self):
        self.testsite  = init_annalist_test_site()
        self.testcoll  = Collection.create(
            self.testsite, "testcoll", collection_create_values("testcoll")
            )
        # Create test type
        self.testtypes = RecordType.create(
            self.testcoll, "testtype", 
            recordtype_create_values(
                coll_id="testcoll", type_id="testtype", type_uri="test:testtype",
                supertype_uris=[]
                )
            )
        # Create test type data parents
        self.testdata = RecordTypeData.create(self.testcoll, "testtype", {})
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        resetSitedata(scope="collections")
        return

    @classmethod
    def setUpClass(cls):
        super(FieldSubpropertyTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(FieldSubpropertyTest, cls).tearDownClass()
        resetSitedata()
        return

    def create_subproperty_field_view_entity(self):
        # Create test field using superproperty
        self.test_sup_field = RecordField.create(self.testcoll, "Test_sup_field",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "Field using superproperty URI"
            , RDFS.CURIE.comment:               "Field using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Text"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sup_field)"
            , ANNAL.CURIE.property_uri:         "test:superprop_uri"
            , ANNAL.CURIE.field_placement:      "small:0,12;medium:0,6"
            })
        self.assertTrue(self.test_sup_field is not None)
        # Create test field using subproperty and declaring superproperty
        self.test_sub_field = RecordField.create(self.testcoll, "Test_sub_field",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "Field using superproperty URI"
            , RDFS.CURIE.comment:               "Field using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Text"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sub_field)"
            , ANNAL.CURIE.property_uri:         "test:subprop_uri"
            , ANNAL.CURIE.superproperty_uri:    [{"@id": "test:superprop_uri"}]
            , ANNAL.CURIE.field_placement:      "small:0,12;medium:0,6"
            })
        self.assertTrue(self.test_sub_field is not None)
        # Create test view using superproperty
        self.test_view = RecordView.create(self.testcoll, "testview",
            { ANNAL.CURIE.type:             "annal:View"
            , ANNAL.CURIE.uri:              "test:view"
            , RDFS.CURIE.label:             "Test view label"
            , RDFS.CURIE.comment:           "Test view comment"
            , ANNAL.CURIE.view_entity_type: "test:testtype"
            , ANNAL.CURIE.view_fields:
              [ { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Entity_id"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              , { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Test_sup_field"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              ]
            })
        self.assertTrue(self.test_view is not None)
        # Create test entity using subproperty
        self.testentity_data = EntityData.create(self.testdata, "testentity", 
            entitydata_create_values(
                "testentity", type_id="testtype",
                type_uri="test:testtype",
                extra_fields={"test:subprop_uri": "Test field value"} 
                )
            )
        self.assertTrue(self.testentity_data is not None)
        return

    def create_subproperty_set_field_view_entity(self):
        # Create set (multi-value) field using superproperty
        self.test_sup_set = RecordField.create(self.testcoll, "Test_sup_set",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "Set using superproperty URI"
            , RDFS.CURIE.comment:               "Set using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Group_Set_Row"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_value_type:     "annal:Test_sup_set"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sup_set)"
            , ANNAL.CURIE.property_uri:         "test:superprop_uri"
            , ANNAL.CURIE.field_placement:      "small:0,12"
            , ANNAL.CURIE.field_fields:
              [ { ANNAL.CURIE.field_id:         "_field/Test_sup_field"
                , ANNAL.CURIE.property_uri:     "@value"
                , ANNAL.CURIE.field_placement:  "small:0,12"
                }
              ]
            , ANNAL.CURIE.repeat_label_add:     "Add sup entity"
            , ANNAL.CURIE.repeat_label_delete:  "Remove sup entity"
            })
        self.assertTrue(self.test_sup_set is not None)
        # Create list field using subproperty and declaring superproperty
        self.test_sub_set = RecordField.create(self.testcoll, "Test_sub_set",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "Set using superproperty URI"
            , RDFS.CURIE.comment:               "Set using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Group_Set_Row"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_value_type:     "annal:Test_sub_set"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sub_set)"
            , ANNAL.CURIE.property_uri:         "test:subprop_set_uri"
            , ANNAL.CURIE.superproperty_uri:    [{"@id": "test:superprop_uri"}]
            , ANNAL.CURIE.field_placement:      "small:0,12"
            , ANNAL.CURIE.field_fields:
              [ { ANNAL.CURIE.field_id:         "_field/Test_sub_field"
                , ANNAL.CURIE.property_uri:     "@value"
                , ANNAL.CURIE.field_placement:  "small:0,12"
                }
              ]
            , ANNAL.CURIE.repeat_label_add:     "Add sub entity"
            , ANNAL.CURIE.repeat_label_delete:  "Remove sub entity"
            })
        self.assertTrue(self.test_sub_set is not None)
        # Create test view using superproperty
        self.test_view = RecordView.create(self.testcoll, "testsetview",
            { ANNAL.CURIE.type:             "annal:View"
            , ANNAL.CURIE.uri:              "test:setview"
            , RDFS.CURIE.label:             "Test setview label"
            , RDFS.CURIE.comment:           "Test setview comment"
            , ANNAL.CURIE.view_entity_type: "test:testtype"
            , ANNAL.CURIE.view_fields:
              [ { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Entity_id"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              , { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Test_sup_set"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              ]
            })
        self.assertTrue(self.test_view is not None)
        # Create test entity using list and item subproperty URIs
        self.testsetentity_data = EntityData.create(self.testdata, "testsetentity", 
            entitydata_create_values(
                "testsetentity", type_id="testtype",
                type_uri="test:testtype",
                extra_fields=
                  { "test:subprop_uri":
                    [ { "@value": "Test field 1 value"}
                    , { "@value": "Test field 2 value"}
                    ]
                  }
                )
            )
        self.assertTrue(self.testsetentity_data is not None)
        return

    def create_subproperty_list_field_view_entity(self):
        # Create list field using superproperty
        self.test_sup_list = RecordField.create(self.testcoll, "Test_sup_list",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "List using superproperty URI"
            , RDFS.CURIE.comment:               "List using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Group_Seq_Row"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_value_type:     "annal:Test_sup_list"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sup_list)"
            , ANNAL.CURIE.property_uri:         "test:superprop_list_uri"
            , ANNAL.CURIE.field_placement:      "small:0,12"
            , ANNAL.CURIE.field_fields:
              [ { ANNAL.CURIE.field_id:         "_field/Test_sup_field"
                , ANNAL.CURIE.field_placement:  "small:0,12"
                }
              ]
            , ANNAL.CURIE.repeat_label_add:     "Add sup entity"
            , ANNAL.CURIE.repeat_label_delete:  "Remove sup entity"
            })
        self.assertTrue(self.test_sup_list is not None)
        # Create list field using subproperty and declaring superproperty
        self.test_sub_list = RecordField.create(self.testcoll, "Test_sub_list",
            { ANNAL.CURIE.type:                 "annal:Field"
            , RDFS.CURIE.label:                 "List using superproperty URI"
            , RDFS.CURIE.comment:               "List using superproperty URI"
            , ANNAL.CURIE.field_render_type:    "_enum_render_type/Group_Seq_Row"
            , ANNAL.CURIE.field_value_mode:     "_enum_value_mode/Value_direct"
            , ANNAL.CURIE.field_value_type:     "annal:Test_sub_list"
            , ANNAL.CURIE.field_entity_type:    "test:testtype"
            , ANNAL.CURIE.placeholder:          "(Test_sub_list)"
            , ANNAL.CURIE.property_uri:         "test:subprop_list_uri"
            , ANNAL.CURIE.superproperty_uri:    [{"@id": "test:superprop_list_uri"}]
            , ANNAL.CURIE.field_placement:      "small:0,12"
            , ANNAL.CURIE.field_fields:
              [ { ANNAL.CURIE.field_id:         "_field/Test_sub_field"
                , ANNAL.CURIE.field_placement:  "small:0,12"
                }
              ]
            , ANNAL.CURIE.repeat_label_add:     "Add sub entity"
            , ANNAL.CURIE.repeat_label_delete:  "Remove sub entity"
            })
        self.assertTrue(self.test_sub_list is not None)
        # Create test view using superproperty
        self.test_view = RecordView.create(self.testcoll, "testlistview",
            { ANNAL.CURIE.type:             "annal:View"
            , ANNAL.CURIE.uri:              "test:listview"
            , RDFS.CURIE.label:             "Test listview label"
            , RDFS.CURIE.comment:           "Test listview comment"
            , ANNAL.CURIE.view_entity_type: "test:testtype"
            , ANNAL.CURIE.view_fields:
              [ { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Entity_id"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              , { ANNAL.CURIE.field_id:         layout.FIELD_TYPEID+"/Test_sup_list"
                , ANNAL.CURIE.field_placement:  "small:0,12;medium:0,6"
                }
              ]
            })
        self.assertTrue(self.test_view is not None)
        # Create test entity using list and item subproperty URIs
        self.testlistentity_data = EntityData.create(self.testdata, "testlistentity", 
            entitydata_create_values(
                "testlistentity", type_id="testtype",
                type_uri="test:testtype",
                extra_fields=
                  { "test:subprop_list_uri":
                    [ { "test:subprop_uri": "Test field 1 value"}
                    , { "test:subprop_uri": "Test field 2 value"}
                    ]
                  }
                )
            )
        self.assertTrue(self.testlistentity_data is not None)
        return

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_view_subproperty_field(self):
        self.create_subproperty_field_view_entity()
        # Render view
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", 
            entity_id="testentity", 
            view_id="testview"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context
        url = entity_url(coll_id="testcoll", type_id="testtype", entity_id="testentity")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "testentity")
        self.assertEqual(r.context['orig_id'],          "testentity")
        self.assertEqual(r.context['action'],           "view")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        # Check id field for sanity
        f1 = context_view_field(r.context, 0, 0)
        self.assertEqual(f1.field_id,           "Entity_id")
        self.assertEqual(f1.description['field_property_uri'], ANNAL.CURIE.id)
        self.assertEqual(f1['field_value'], self.testentity_data[ANNAL.CURIE.id])
        # Check superproperty field value
        f2 = context_view_field(r.context, 1, 0)
        self.assertEqual(f2.field_id,           "Test_sup_field")
        self.assertEqual(f2.description['field_property_uri'], "test:superprop_uri")
        self.assertEqual(f2['field_value'], self.testentity_data["test:subprop_uri"])
        return

    # def test_list_field_alias(self):
    #     # List BibEntry fields in Default_list
    #     u = entitydata_list_type_url("testcoll", "testtype", list_id="Default_list")
    #     r = self.client.get(u)
    #     self.assertEqual(r.status_code,   200)
    #     self.assertEqual(r.reason_phrase, "OK")
    #     # log.info(r.content) #@@
    #     # Test context
    #     self.assertEqual(r.context['coll_id'],  "testcoll")
    #     self.assertEqual(r.context['type_id'],  "testtype")
    #     self.assertEqual(r.context['list_choices']['field_value'], "Default_list")
    #     # Fields
    #     head_fields = context_list_head_fields(r.context)
    #     self.assertEqual(len(head_fields), 1)       # One row of 2 cols..
    #     self.assertEqual(len(head_fields[0]['row_field_descs']), 2)
    #     f0 = context_view_field(r.context, 0, 0)
    #     f1 = context_view_field(r.context, 0, 1)
    #     # 1st field
    #     self.assertEqual(f0['field_id'], 'Entity_id')
    #     self.assertEqual(f0['field_property_uri'], "annal:id")
    #     self.assertEqual(f0['field_value'], "")
    #     # 2nd field
    #     self.assertEqual(f1['field_id'], 'Entity_label')
    #     self.assertEqual(f1['field_property_uri'], "rdfs:label")
    #     self.assertEqual(f1['field_value'], "")
    #     # List entities (actually, just the one)
    #     entities = context_list_entities(r.context)
    #     self.assertEqual(len(entities), 1)
    #     self.assertEqual(
    #         context_list_item_field_value(r.context, entities[0], 0), 
    #         "testentity"
    #         )
    #     self.assertEqual(
    #         context_list_item_field_value(r.context, entities[0], 1), 
    #         self.testentity_data['bib:title']
    #         )
    #     return

    def test_save_subproperty_field(self):
        self.create_subproperty_field_view_entity()
        # Post edit form response
        u = entitydata_edit_url(
            "edit", "testcoll", "testtype", 
            entity_id="testentity", 
            view_id="testview"
            )
        f = (
            { 'entity_id':          "testentity"
            , 'entity_type':        "testtype"
            , 'orig_id':            "testentity"
            , 'orig_type':          "testtype"
            , 'Test_sup_field':     "Updated subproperty value"
            , 'action':             "edit"
            , 'save':               "Save"
            })
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        # Check entity exists,and compare data with expected
        typeinfo = EntityTypeInfo(self.testcoll, "testtype")
        self.assertTrue(typeinfo.entityclass.exists(typeinfo.entityparent, "testentity"))
        e = typeinfo.entityclass.load(typeinfo.entityparent, "testentity")
        self.assertEqual(e.get_id(), "testentity")
        # Check superproperty value remains undefined
        self.assertEqual(e.get_values().get("test:superprop_uri", "undefined"), "undefined")
        # Check subproperty has been updated
        v = self.testentity_data.get_values().copy()
        v['test:subprop_uri']   = f['Test_sup_field']
        self.assertDictionaryMatch(e.get_values(), v)
        return

    def test_view_subproperty_set(self):
        self.create_subproperty_field_view_entity()
        self.create_subproperty_set_field_view_entity()
        # Post edit form response
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", 
            entity_id="testsetentity", 
            view_id="testsetview"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context
        url = entity_url(coll_id="testcoll", type_id="testtype", entity_id="testsetentity")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "testsetentity")
        self.assertEqual(r.context['orig_id'],          "testsetentity")
        self.assertEqual(r.context['action'],           "view")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        # Check id field for sanity
        f1 = context_view_field(r.context, 0, 0)
        self.assertEqual(f1.field_id,           "Entity_id")
        self.assertEqual(f1.description['field_property_uri'], ANNAL.CURIE.id)
        self.assertEqual(f1['field_value'], self.testsetentity_data[ANNAL.CURIE.id])
        # Check superproperty field value
        f2    = context_view_field(r.context, 1, 0)
        self.assertEqual(f2.field_id,           "Test_sup_set")
        self.assertEqual(f2.description['field_property_uri'], "test:superprop_uri")
        f2rfs = context_view_repeat_fields(r.context, f2)
        self.assertEqual(f2rfs[0][0].field_value, "Test field 1 value")
        self.assertEqual(f2rfs[1][0].field_value, "Test field 2 value")
        # Check rendered list
        item1data = ("""
            <div class="row view-value-col">
              <div class="view-value small-12 columns">
                <span>Test field 1 value</span>
              </div>
            </div>
            """)
        self.assertContains(r, item1data, html=True)
        item2data = ("""
            <div class="row view-value-col">
              <div class="view-value small-12 columns">
                <span>Test field 1 value</span>
              </div>
            </div>
            """)
        self.assertContains(r, item2data, html=True)
        return

    def test_save_subproperty_set(self):
        self.create_subproperty_field_view_entity()
        self.create_subproperty_set_field_view_entity()
        # Post edit form response
        u = entitydata_edit_url(
            "edit", "testcoll", "testtype", 
            entity_id="testsetentity", 
            view_id="testsetview"
            )
        f = (
            { 'entity_id':          "testsetentity"
            , 'entity_type':        "testtype"
            , 'orig_id':            "testsetentity"
            , 'orig_type':          "testtype"
            , 'Test_sup_set__0__Test_sup_field': "Updated subprop 1"
            , 'Test_sup_set__1__Test_sup_field': "Updated subprop 2"
            , 'action':             "edit"
            , 'save':               "Save"
            })
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        # Check entity exists,and compare data with expected
        typeinfo = EntityTypeInfo(self.testcoll, "testtype")
        self.assertTrue(typeinfo.entityclass.exists(typeinfo.entityparent, "testsetentity"))
        e = typeinfo.entityclass.load(typeinfo.entityparent, "testsetentity")
        self.assertEqual(e.get_id(), "testsetentity")
        # Check superproperty value remains undefined
        self.assertEqual(e.get_values().get("test:superprop_uri", "undefined"), "undefined")
        # Check subproperty has been updated
        v = self.testsetentity_data.get_values().copy()
        v['test:subprop_uri'] = (
            [ { "@value": "Updated subprop 1" }
            , { "@value": "Updated subprop 2" }
            ])
        self.assertDictionaryMatch(e.get_values(), v)
        return

    def test_view_subproperty_list(self):
        self.create_subproperty_field_view_entity()
        self.create_subproperty_list_field_view_entity()
        # Post edit form response
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", 
            entity_id="testlistentity", 
            view_id="testlistview"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context
        url = entity_url(coll_id="testcoll", type_id="testtype", entity_id="testlistentity")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "testlistentity")
        self.assertEqual(r.context['orig_id'],          "testlistentity")
        self.assertEqual(r.context['action'],           "view")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        # Check id field for sanity
        f1 = context_view_field(r.context, 0, 0)
        self.assertEqual(f1.field_id,           "Entity_id")
        self.assertEqual(f1.description['field_property_uri'], ANNAL.CURIE.id)
        self.assertEqual(f1['field_value'], self.testlistentity_data[ANNAL.CURIE.id])
        # Check superproperty field value
        f2    = context_view_field(r.context, 1, 0)
        self.assertEqual(f2.field_id,           "Test_sup_list")
        self.assertEqual(f2.description['field_property_uri'], "test:superprop_list_uri")
        f2rfs = context_view_repeat_fields(r.context, f2)
        self.assertEqual(f2rfs[0][0].field_value, "Test field 1 value")
        self.assertEqual(f2rfs[1][0].field_value, "Test field 2 value")
        # Check rendered list
        item1data = ("""
            <div class="row view-value-col">
              <div class="view-value small-12 columns">
                <span>Test field 1 value</span>
              </div>
            </div>
            """)
        self.assertContains(r, item1data, html=True)
        item2data = ("""
            <div class="row view-value-col">
              <div class="view-value small-12 columns">
                <span>Test field 1 value</span>
              </div>
            </div>
            """)
        self.assertContains(r, item2data, html=True)
        return

    def test_save_subproperty_list(self):
        self.create_subproperty_field_view_entity()
        self.create_subproperty_list_field_view_entity()
        # Post edit form response
        u = entitydata_edit_url(
            "edit", "testcoll", "testtype", 
            entity_id="testlistentity", 
            view_id="testlistview"
            )
        f = (
            { 'entity_id':          "testlistentity"
            , 'entity_type':        "testtype"
            , 'orig_id':            "testlistentity"
            , 'orig_type':          "testtype"
            , 'Test_sup_list__0__Test_sup_field': "Updated subprop 1"
            , 'Test_sup_list__1__Test_sup_field': "Updated subprop 2"
            , 'action':             "edit"
            , 'save':               "Save"
            })
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        # Check entity exists,and compare data with expected
        typeinfo = EntityTypeInfo(self.testcoll, "testtype")
        self.assertTrue(typeinfo.entityclass.exists(typeinfo.entityparent, "testlistentity"))
        e = typeinfo.entityclass.load(typeinfo.entityparent, "testlistentity")
        # print("@@@@ e: "+repr(e.get_values()))
        self.assertEqual(e.get_id(), "testlistentity")
        # Check superproperty value remains undefined
        self.assertEqual(e.get_values().get("test:superprop_uri", "undefined"), "undefined")
        # Check subproperty has been updated
        v = self.testlistentity_data.get_values().copy()
        #@@NOTE:
        #   With reference to 'test:superprop_uri' below:
        #   It is a known restriction that subproperties used within repeating field
        #   groups are not propagated.  See RepeatValuesMap.map_form_to_entity.
        #@@
        v['test:subprop_list_uri'] = (
            [ { "test:superprop_uri": "Updated subprop 1" }
            , { "test:superprop_uri": "Updated subprop 2" }
            ])
        self.assertDictionaryMatch(e.get_values(), v)
        return

# End.



"""
Tests for field placement (size/position) rendering
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os
import unittest
import re

from annalist.views.fields.render_placement import (
    Placement, LayoutOptions, get_placement_classes,
    get_field_placement_renderer, 
    option_body, view_body
    )

from .field_rendering_support import FieldRendererTestSupport

class FieldPlacementRenderingTest(FieldRendererTestSupport):

    def setUp(self):
        self.placement_context = self._make_test_context("small:0,12;medium:4,4")
        return

    def tearDown(self):
        return

    def test_get_placement_classes(self):
        """
        This was a DocTest, but Py3 compatibility....
        """
        self.assertEqual(
            get_placement_classes("medium:0,12"),
            Placement(
                width=LayoutOptions(s=12, m=12, l=12),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-12 columns',
                label='small-12 medium-2 columns',
                value='small-12 medium-10 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("large:0,12"),
            Placement(
                width=LayoutOptions(s=12, m=12, l=12),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-12 columns',
                label='small-12 medium-2 columns',
                value='small-12 medium-10 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,12;medium:0,4"),
            Placement(
                width=LayoutOptions(s=12, m=4, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-12 medium-4 columns',
                label='small-12 medium-6 columns',
                value='small-12 medium-6 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,12; medium:0,4"),
            Placement(
                width=LayoutOptions(s=12, m=4, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-12 medium-4 columns',
                label='small-12 medium-6 columns',
                value='small-12 medium-6 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,12;medium:0,6;large:0,4"),
            Placement(
                width=LayoutOptions(s=12, m=6, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-12 medium-6 large-4 columns',
                label='small-12 medium-4 large-6 columns',
                value='small-12 medium-8 large-6 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,6;medium:0,4"),
            Placement(
                width=LayoutOptions(s=6, m=4, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-6 medium-4 columns',
                label='small-12 medium-6 columns',
                value='small-12 medium-6 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,6;medium:0,4,right"),
            Placement(
                width=LayoutOptions(s=6, m=4, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-6 medium-4 columns right',
                label='small-12 medium-6 columns',
                value='small-12 medium-6 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,6"),
            Placement(
                width=LayoutOptions(s=6, m=6, l=6),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=True, m=True, l=True),
                field='small-6 columns',
                label='small-12 medium-4 columns',
                value='small-12 medium-8 columns'
                )
            )
        self.assertEqual(
            get_placement_classes("small:0,12,hide;medium:0,4"),
            Placement(
                width=LayoutOptions(s=12, m=4, l=4),
                offset=LayoutOptions(s=0, m=0, l=0),
                display=LayoutOptions(s=False, m=True, l=True),
                field='small-12 medium-4 columns show-for-medium-up',
                label='small-12 medium-6 columns',
                value='small-12 medium-6 columns'
                )
            )
        return

    def test_RenderFieldPlacementValue(self):
        expect_rendered_view=view_body(
            '''<span class="placement-text">'''+
                '''....####.... (4/4)'''+
                '''</span>'''
                )
        expect_rendered_edit=option_body(
            '''<select class="placement-text" name="repeat_prefix_test_field">\n'''+
            '''  <option value="">(test placeholder)</option>\n'''+
            '''  <option value="small:0,12">'''+
                '''############ (0/12)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:0,6">'''+
                '''######...... (0/6)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:3,6">'''+
                '''...######... (3/6)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:6,6">'''+
                '''......###### (6/6)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:0,4">'''+
                '''####........ (0/4)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:4,4" selected="selected">'''+
                '''....####.... (4/4)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:8,4">'''+
                '''........#### (8/4)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:0,3">'''+
                '''###......... (0/3)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:3,3">'''+
                '''...###...... (3/3)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:6,3">'''+
                '''......###... (6/3)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:9,3">'''+
                '''.........### (9/3)'''+
                '''</option>\n'''+
            # Old right-aligned placement strings
            '''  <option value="small:0,12;medium:6,6right">'''+
                '''......###### (6/6R)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:8,4right">'''+
                '''........#### (8/4R)'''+
                '''</option>\n'''+
            '''  <option value="small:0,12;medium:9,3right">'''+
                '''.........### (9/3R)'''+
                '''</option>\n'''+
            # Column placement strings affecting all display sizes
            '''  <option value="small:0,9">'''+
                '''#########... (0/9col)'''+
                '''</option>\n'''+
            '''  <option value="small:3,9">'''+
                '''...######### (3/9col)'''+
                '''</option>\n'''+
            '''  <option value="small:0,8">'''+
                '''########.... (0/8col)'''+
                '''</option>\n'''+
            '''  <option value="small:4,8">'''+
                '''....######## (4/8col)'''+
                '''</option>\n'''+
            '''  <option value="small:0,6">'''+
                '''######...... (0/6col)'''+
                '''</option>\n'''+
            '''  <option value="small:3,6">'''+
                '''...######... (3/6col)'''+
                '''</option>\n'''+
            '''  <option value="small:6,6">'''+
                '''......###### (6/6col)'''+
                '''</option>\n'''+
            '''  <option value="small:0,4">'''+
                '''####........ (0/4col)'''+
                '''</option>\n'''+
            '''  <option value="small:4,4">'''+
                '''....####.... (4/4col)'''+
                '''</option>\n'''+
            '''  <option value="small:8,4">'''+
                '''........#### (8/4col)'''+
                '''</option>\n'''+
            '''  <option value="small:0,3">'''+
                '''###......... (0/3col)'''+
                '''</option>\n'''+
            '''  <option value="small:3,3">'''+
                '''...###...... (3/3col)'''+
                '''</option>\n'''+
            '''  <option value="small:6,3">'''+
                '''......###... (6/3col)'''+
                '''</option>\n'''+
            '''  <option value="small:9,3">'''+
                '''.........### (9/3col)'''+
                '''</option>\n'''+
            '''</select>'''
            )
        self._check_value_renderer_results(
            get_field_placement_renderer(),
            context=self.placement_context,
            expect_rendered_view=expect_rendered_view,
            expect_rendered_edit=expect_rendered_edit
            )
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()
